@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -37,6 +37,20 @@ images = { }
 # tag.
 image_attributes = collections.defaultdict(list)
 
+# The set of image names Ren'Py knows about, as strings with spaces.
+image_names = [ ]
+
+
+def list_images():
+    """
+    :doc: image_func
+
+    Returns a list of images that have been added to Ren'Py, as a list of
+    strings with spaces between the name components.
+    """
+
+    return image_names
+
 
 def get_available_image_tags():
     """
@@ -47,9 +61,10 @@ def get_available_image_tags():
 
     return [ k for k, v in image_attributes.items() if v ]
 
+
 def get_available_image_attributes(tag, attributes=()):
     """
-    :doc: image_func
+    :undocumented:
 
     Returns a list of tuples, with each tuple representing a possible
     combination of image attributes that can be associated with `tag`.
@@ -72,6 +87,94 @@ def get_available_image_attributes(tag, attributes=()):
     return rv
 
 
+def get_tag_method(tag, method):
+    """
+    This looks for an image named `tag`, then tries to grab the given method
+    from it. If it can do that, returns a reference to the method. Otherwise,
+    it returns None.
+    """
+
+    ti =  images.get((tag,), None)
+
+    if ti is None:
+        return None
+
+    return getattr(ti, method, None)
+
+
+def check_image_attributes(tag, attributes):
+    """
+    :doc: image_func
+
+    Checks to see if there is a unique image with the given tag and
+    attributes. If there is, returns the attributes in order.
+    Otherwise, returns None.
+    """
+
+    ca = get_tag_method(tag, "_choose_attributes")
+
+    if ca is not None:
+        return ca(tag, attributes, None)
+
+    l = get_available_image_attributes(tag, attributes)
+
+    # Check to see if there's an image that is exactly the one we want.
+    for i in l:
+        if len(i) == len(attributes):
+            return tuple(i)
+
+    if len(l) != 1:
+        return None
+
+    return tuple(l[0])
+
+
+def get_ordered_image_attributes(tag, attributes=(), sort=None):
+    """
+    :doc: image_func
+
+    Returns a list of image tags, ordered in a way that makes sense to
+    present to the user.
+
+    `attributes`
+        If present, only attributes that are compatible with the given
+        attributes are considered. (Compatible means that the attributes
+        can be in a single image at the same time.)
+
+    `sort`
+        If not None, the returned list of attributes is sorted. This is a function
+        that should be used as a tiebreaker.
+
+    """
+
+    la = get_tag_method(tag, "_list_attributes")
+
+    if la is not None:
+        return la(tag, attributes)
+
+    attrcount = collections.defaultdict(int)
+    attrtotalpos = collections.defaultdict(float)
+
+    for attrlist in get_available_image_attributes(tag, attributes):
+        for i, attr in enumerate(attrlist):
+            attrcount[attr] += 1
+            attrtotalpos[attr] += i
+
+    if sort is None:
+
+        return list(attrcount.keys())
+
+    else:
+
+        l = [ ]
+
+        for attr in attrcount:
+            l.append((attrtotalpos[attr] / attrcount[attr], sort(attr), attr))
+
+        l.sort()
+        return [ i[2] for i in l ]
+
+
 def register_image(name, d):
     """
     Registers the existence of an image with `name`, and that the image
@@ -87,6 +190,8 @@ def register_image(name, d):
     images[name] = d
     image_attributes[tag].append(rest)
 
+    image_names.append(" ".join(name))
+
 
 def image_exists(name, exact=False):
     """
@@ -135,7 +240,17 @@ class ImageReference(renpy.display.core.Displayable):
     """
 
     nosave = [ 'target' ]
+
     target = None
+    old_transform = None
+    param_target = None
+
+    __version__ = 1
+
+    def after_upgrade(self, version):
+        if version < 1:
+            if isinstance(self.param_target, renpy.display.transform.Transform):
+                self.old_transform = self.param_target
 
     def __init__(self, name, **properties):
         """
@@ -166,6 +281,7 @@ class ImageReference(renpy.display.core.Displayable):
         return True
 
     def _target(self):
+
         if self.target is None:
             self.find_target()
 
@@ -206,15 +322,28 @@ class ImageReference(renpy.display.core.Displayable):
         try:
 
             a = self._args.copy(name=name, args=args)
+
             self.target = target._duplicate(a)
 
-        except Exception, e:
+        except Exception as e:
 
             if renpy.config.debug:
                 raise
 
             error(str(e))
 
+        # Copy the old transform over.
+        new_transform = self.target._target()
+
+        if isinstance(new_transform, renpy.display.transform.Transform):
+            if self.old_transform is not None:
+                new_transform.take_state(self.old_transform)
+
+            self.old_transform = new_transform
+
+        else:
+            self.old_transform = None
+
         return True
 
     _duplicatable = True
@@ -228,8 +357,38 @@ class ImageReference(renpy.display.core.Displayable):
             if rv.name._duplicatable:
                 rv.name = rv.name._duplicate(args)
 
+        rv.find_target()
+        rv._duplicatable = rv.target._duplicatable
+
         return rv
 
+    def _unique(self):
+
+        if self.target is None:
+            self.find_target()
+
+        self._duplicatable = self.target._duplicatable
+
+    def _in_current_store(self):
+
+        if self.target is None:
+            self.find_target()
+
+        target = self.target._in_current_store()
+
+        if target is self.target:
+            return self
+
+        rv = self._copy()
+        rv.target = target
+        return rv
+
+    def _handles_event(self, event):
+        if self.target is None:
+            return False
+
+        return self.target._handles_event(event)
+
     def _hide(self, st, at, kind):
         if self.target is None:
             self.find_target()
@@ -302,6 +461,12 @@ class DynamicImage(renpy.display.core.Displayable):
     # The raw target that the image resolves to, before it has been parameterized.
     raw_target = None
 
+    # Have we been locked, so we never change?
+    locked = False
+
+    # The name used for hashing.
+    hash_name = None
+
     def __init__(self, name, scope=None, **properties):
         super(DynamicImage, self).__init__(**properties)
 
@@ -317,8 +482,9 @@ class DynamicImage(renpy.display.core.Displayable):
             self._duplicatable = True
 
         if isinstance(name, list):
-            self._duplicatable = True
-
+            for i in name:
+                if ("[prefix_" in i):
+                    self._duplicatable = True
 
     def _scope(self, scope, update):
         return self.find_target(scope, update)
@@ -327,7 +493,14 @@ class DynamicImage(renpy.display.core.Displayable):
         return u"DynamicImage {!r}".format(self.name)
 
     def __hash__(self):
-        return hash(self.name)
+
+        if self.hash_name is None:
+            if isinstance(self.name, list):
+                self.hash_name = tuple(self.name)
+            else:
+                self.hash_name = self.name
+
+        return hash(self.hash_name)
 
     def __eq__(self, o):
         if self is o:
@@ -349,22 +522,31 @@ class DynamicImage(renpy.display.core.Displayable):
 
     def find_target(self, scope=None, update=True):
 
+        if self.locked and (self.target is not None):
+            return
+
         if self._args.prefix is None:
             prefix = ""
         else:
             prefix = self._args.prefix
 
         try:
-            target = renpy.easy.dynamic_image(self.name, scope, prefix=prefix)
+            search = [ ]
+            target = renpy.easy.dynamic_image(self.name, scope, prefix=prefix, search=search)
+        except KeyError as ke:
+            raise Exception("In DynamicImage %r: Could not find substitution '%s'." % (self.name, unicode(ke.args[0])))
         except Exception as e:
-            raise
             raise Exception("In DynamicImage %r: %r" % (self.name, e))
 
         if target is None:
-            error = "DynamicImage %r: did not resolve to an image." % (self.name,)
+            error = "DynamicImage %r: could not find image." % (self.name,)
 
-            if self._args.prefix:
-                error += " prefix=" + self._args.prefix
+            if len(search) == 1:
+                error += " (%r)" % (search[0],)
+            elif len(search) == 2:
+                error += " (%r, %r)" % (search[0], search[1])
+            elif len(search) > 2:
+                error += " (%r, %r, and %d more.)" % (search[0], search[1], len(search) - 2)
 
             raise Exception(error)
 
@@ -400,8 +582,25 @@ class DynamicImage(renpy.display.core.Displayable):
     def _duplicate(self, args):
         rv = self._copy(args)
         rv.target = None
+        # This does not set _duplicatable, since it should always remain the
+        # same.
         return rv
 
+    def _in_current_store(self):
+        rv = self._copy()
+
+        if rv.target:
+            rv.target = rv.target._in_current_store()
+
+        rv.locked = True
+        return rv
+
+    def _handles_event(self, event):
+        if self.target is None:
+            return False
+
+        return self.target._handles_event(event)
+
     def _hide(self, st, at, kind):
         if self.target is None:
             self.find_target()
@@ -480,7 +679,6 @@ class ShownImageInfo(renpy.object.Object):
             self.attributes = old.attributes.copy()
             self.shown = old.shown.copy()
 
-
     def after_upgrade(self, version):
         if version < 2:
 
@@ -492,13 +690,13 @@ class ShownImageInfo(renpy.object.Object):
                     self.attributes[layer, tag] = self.images[layer][tag][1:]
                     self.shown.add((layer, tag))
 
-    def get_attributes(self, layer, tag):
+    def get_attributes(self, layer, tag, default=()):
         """
         Get the attributes associated the image with tag on the given
         layer.
         """
 
-        return self.attributes.get((layer, tag), ())
+        return self.attributes.get((layer, tag), default)
 
     def showing(self, layer, name, exact=False):
         """
@@ -582,7 +780,9 @@ class ShownImageInfo(renpy.object.Object):
 
         # If the name matches one that exactly exists, return it.
         if (name in images) and not (wanted or remove):
-            return name
+            ca = get_tag_method(tag, "_choose_attributes")
+            if ca is None:
+                return name
 
         nametag = name[0]
 
@@ -601,11 +801,17 @@ class ShownImageInfo(renpy.object.Object):
         for i in remove:
             optional.discard(i)
 
-
         return self.choose_image(nametag, required, optional, name)
 
     def choose_image(self, tag, required, optional, exception_name):
 
+        ca = get_tag_method(tag, "_choose_attributes")
+        if ca is not None:
+            attrs = ca(tag, required, optional)
+
+            if attrs is not None:
+                return (tag,) + attrs
+
         # The longest length of an image that matches.
         max_len = -1
 
@@ -647,13 +853,28 @@ class ShownImageInfo(renpy.object.Object):
             return None
 
         if len(matches) == 1:
-            return matches[0]
+
+            rv = matches[0]
+
+            d = images.get(rv, None)
+            ca = getattr(d, "_choose_attributes", None)
+
+            if ca is not None:
+
+                required = set()
+                optional = set(i for i in optional if i not in rv)
+                tag = " ".join(rv)
+
+                rv =  rv + ca(tag, required, optional)
+
+            return rv
 
         if exception_name:
             raise Exception("Showing '" + " ".join(exception_name) + "' is ambiguous, possible images include: " + ", ".join(" ".join(i) for i in matches))
         else:
             return None
 
+
 renpy.display.core.ImagePredictInfo = ShownImageInfo
 
 
@@ -663,4 +884,3 @@ Image = renpy.display.im.image
 Solid = renpy.display.imagelike.Solid
 Frame = renpy.display.imagelike.Frame
 ImageButton = renpy.display.behavior.ImageButton
-
