@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -103,6 +103,13 @@ class Container(renpy.display.core.Displayable):
 
         super(Container, self).__init__(**properties)
 
+    def _handles_event(self, event):
+        for i in self.children:
+            if i._handles_event(event):
+                return True
+
+        return False
+
     def set_style_prefix(self, prefix, root):
         super(Container, self).set_style_prefix(prefix, root)
 
@@ -252,8 +259,9 @@ class Container(renpy.display.core.Displayable):
         return False
 
 
-def LiveComposite(size, *args, **properties):
+def Composite(size, *args, **properties):
     """
+    :name: Composite
     :doc: disp_imagelike
 
     This creates a new displayable of `size`, by compositing other
@@ -269,7 +277,7 @@ def LiveComposite(size, *args, **properties):
 
     ::
 
-       image eileen composite = LiveComposite(
+       image eileen composite = Composite(
            (300, 600),
            (0, 0), "body.png",
            (0, 0), "clothes.png",
@@ -292,8 +300,13 @@ def LiveComposite(size, *args, **properties):
     return rv
 
 
+LiveComposite = Composite
+
+
 class Position(Container):
     """
+    :undocumented:
+
     Controls the placement of a displayable on the screen, using
     supplied position properties. This is the non-curried form of
     Position, which should be used when the user has directly created
@@ -494,6 +507,7 @@ class MultiBox(Container):
     layer_name = None
     first = True
     order_reverse = False
+    layout = None
 
     def __init__(self, spacing=None, layout=None, style='default', **properties):
 
@@ -616,6 +630,16 @@ class MultiBox(Container):
 
         self.scene_list.extend(l)
 
+    def update_times(self):
+
+        it = renpy.game.interface.interact_time
+
+        self.start_times = [ i or it for i in self.start_times ]
+        self.anim_times = [ i or it for i in self.anim_times ]
+
+        if it is None:
+            self.first = True
+
     def render(self, width, height, st, at):
 
         # Do we need to adjust the child times due to our being a layer?
@@ -637,28 +661,19 @@ class MultiBox(Container):
             self.first = False
 
             if adjust_times:
+                self.update_times()
 
-                it = renpy.game.interface.interact_time
+        layout = self.style.box_layout
 
-                self.start_times = [ i or it for i in self.start_times ]
-                self.anim_times = [ i or it for i in self.anim_times ]
-
-            layout = self.style.box_layout
-
-            if layout is None:
-                layout = self.default_layout
-
-            self.layout = layout  # W0201
-
-        else:
-            layout = self.layout
+        if layout is None:
+            layout = self.default_layout
 
         # Handle time adjustment, store the results in csts and cats.
         if adjust_times:
             t = renpy.game.interface.frame_time
 
-            csts = [ t - start for start in self.start_times ]
-            cats = [ t - anim for anim in self.anim_times ]
+            csts = [ 0 if (start is None) else (t - start) for start in self.start_times ]
+            cats = [ 0 if (anim is None) else (t - anim) for anim in self.anim_times ]
 
         else:
             csts = [ st ] * len(self.children)
@@ -952,6 +967,14 @@ class MultiBox(Container):
 
     def event(self, ev, x, y, st):
 
+        # Do we need to adjust the child times due to our being a layer?
+        if self.first:
+
+            self.first = False
+
+            if self.layer_name or (self.layers is not None):
+                self.update_times()
+
         children_offsets = zip(self.children, self.offsets, self.start_times)
 
         if not self.style.order_reverse:
@@ -1242,8 +1265,13 @@ class DynamicDisplayable(renpy.display.core.Displayable):
             else:
                 child, _ = self.function(0, 0, *self.args, **self.kwargs)
 
-            if child is not None:
+            if isinstance(child, list):
+
+                for i in child:
+                    renpy.display.predict.displayable(i)
+            else:
                 renpy.display.predict.displayable(child)
+
         except:
             pass
 
@@ -1283,13 +1311,16 @@ def condition_switch_pick(switch):
     raise Exception("Switch could not choose a displayable.")
 
 
-def condition_switch_show(st, at, switch):
+def condition_switch_show(st, at, switch, predict_all=None):
     return condition_switch_pick(switch), None
 
 
-def condition_switch_predict(switch):
+def condition_switch_predict(switch, predict_all=None):
 
-    if renpy.game.lint:
+    if predict_all is None:
+        predict_all = renpy.config.conditionswitch_predict_all
+
+    if renpy.game.lint or (predict_all and renpy.display.predict.predicting):
         return [ d for _cond, d in switch ]
 
     return [ condition_switch_pick(switch) ]
@@ -1298,17 +1329,26 @@ def condition_switch_predict(switch):
 def ConditionSwitch(*args, **kwargs):
     """
     :doc: disp_dynamic
+    :args: (*args, predict_all=None, **properties)
 
     This is a displayable that changes what it is showing based on
-    python conditions. The positional argument should be given in
+    Python conditions. The positional arguments should be given in
     groups of two, where each group consists of:
 
-    * A string containing a python condition.
+    * A string containing a Python condition.
     * A displayable to use if the condition is true.
 
     The first true condition has its displayable shown, at least
     one condition should always be true.
 
+    The conditions uses here should not have externally-visible side-effects.
+
+    `predict_all`
+        If True, all of the possible displayables will be predicted when
+        the displayable is shown. If False, only the current condition is
+        predicted. If None, :var:`config.conditionswitch_predict_all` is
+        used.
+
     ::
 
         image jill = ConditionSwitch(
@@ -1316,6 +1356,7 @@ def ConditionSwitch(*args, **kwargs):
             "True", "jill_sober.png")
     """
 
+    predict_all = kwargs.pop("predict_all", None)
     kwargs.setdefault('style', 'default')
 
     switch = [ ]
@@ -1334,6 +1375,7 @@ def ConditionSwitch(*args, **kwargs):
 
     rv = DynamicDisplayable(condition_switch_show,
                             switch,
+                            predict_all,
                             _predict_function=condition_switch_predict)
 
     return Position(rv, **kwargs)
@@ -1342,6 +1384,7 @@ def ConditionSwitch(*args, **kwargs):
 def ShowingSwitch(*args, **kwargs):
     """
     :doc: disp_dynamic
+    :args: (*args, predict_all=None, **properties)
 
     This is a displayable that changes what it is showing based on the
     images are showing on the screen. The positional argument should
@@ -1352,6 +1395,12 @@ def ShowingSwitch(*args, **kwargs):
 
     A default image should be specified.
 
+    `predict_all`
+        If True, all of the possible displayables will be predicted when
+        the displayable is shown. If False, only the current condition is
+        predicted. If None, :var:`config.conditionswitch_predict_all` is
+        used.
+
     One use of ShowingSwitch is to have side images change depending on
     the current emotion of a character. For example::
 
@@ -1407,19 +1456,23 @@ class IgnoresEvents(Container):
         return None
 
 
-def LiveCrop(rect, child, **properties):
+def Crop(rect, child, **properties):
     """
     :doc: disp_imagelike
+    :name: Crop
 
     This created a displayable by cropping `child` to `rect`, where
     `rect` is an (x, y, width, height) tuple. ::
 
-        image eileen cropped = LiveCrop((0, 0, 300, 300), "eileen happy")
+        image eileen cropped = Crop((0, 0, 300, 300), "eileen happy")
     """
 
     return renpy.display.motion.Transform(child, crop=rect, **properties)
 
 
+LiveCrop = Crop
+
+
 class Side(Container):
 
     possible_positions = set([ 'tl', 't', 'tr', 'r', 'br', 'b', 'bl', 'l', 'c'])
@@ -1653,16 +1706,31 @@ class AdjustTimes(Container):
 
         self.add(child)
 
+    def adjusted_times(self):
+
+        interact_time = renpy.game.interface.interact_time
+
+        if (self.start_time is None) and (interact_time is not None):
+            self.start_time = interact_time
+
+        if self.start_time is not None:
+            st = renpy.game.interface.frame_time - self.start_time
+        else:
+            st = 0
+
+        if (self.anim_time is None) and (interact_time is not None):
+            self.anim_time = interact_time
+
+        if self.anim_time is not None:
+            at = renpy.game.interface.frame_time - self.anim_time
+        else:
+            at = 0
+
+        return st, at
+
     def render(self, w, h, st, at):
 
-        if self.start_time is None:
-            self.start_time = renpy.game.interface.frame_time
-
-        if self.anim_time is None:
-            self.anim_time = renpy.game.interface.frame_time
-
-        st = renpy.game.interface.frame_time - self.start_time
-        at = renpy.game.interface.frame_time - self.anim_time
+        st, at = self.adjusted_times()
 
         cr = renpy.display.render.render(self.child, w, h, st, at)
         cw, ch = cr.get_size()
@@ -1673,19 +1741,24 @@ class AdjustTimes(Container):
 
         return rv
 
+    def event(self, ev, x, y, st):
+        st, _ = self.adjusted_times()
+        Container.event(self, ev, x, y, st)
+
     def get_placement(self):
         return self.child.get_placement()
 
 
-class LiveTile(Container):
+class Tile(Container):
     """
     :doc: disp_imagelike
+    :name: Tile
 
     Tiles `child` until it fills the area allocated to this displayable.
 
     ::
 
-        image bg tile = LiveTile("bg.png")
+        image bg tile = Tile("bg.png")
 
     """
 
@@ -1721,6 +1794,9 @@ class LiveTile(Container):
         return rv
 
 
+LiveTile = Tile
+
+
 class Flatten(Container):
     """
     :doc: disp_imagelike
@@ -1771,11 +1847,7 @@ class AlphaMask(Container):
     opaque where `child` and `mask` are both opaque.
 
     The `child` and `mask` parameters may be arbitrary displayables. The
-    size of the AlphaMask is the size of the overlap between `child` and
-    `mask`.
-
-    Note that this takes different arguments from :func:`im.AlphaMask`,
-    which uses the mask's color channel.
+    size of the AlphaMask is the size of `child`.
     """
 
     def __init__(self, child, mask, **properties):
@@ -1784,24 +1856,19 @@ class AlphaMask(Container):
         self.add(child)
         self.mask = renpy.easy.displayable(mask)
         self.null = None
-        self.size = None
 
     def render(self, width, height, st, at):
 
         cr = renpy.display.render.render(self.child, width, height, st, at)
-        mr = renpy.display.render.render(self.mask, width, height, st, at)
+        w, h = cr.get_size()
 
-        cw, ch = cr.get_size()
-        mw, mh = mr.get_size()
+        mr = renpy.display.render.Render(w, h)
+        mr.place(self.mask, main=False)
 
-        w = min(cw, mw)
-        h = min(ch, mh)
-        size = (w, h)
+        if self.null is None:
+            self.null = Fixed()
 
-        if self.size != size:
-            self.null = Null(w, h)
-
-        nr = renpy.display.render.render(self.null, width, height, st, at)
+        nr = renpy.display.render.render(self.null, w, h, st, at)
 
         rv = renpy.display.render.Render(w, h, opaque=False)
 
