@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -35,7 +35,6 @@ import marshal
 import random
 import weakref
 import re
-import sets
 import sys
 import time
 
@@ -53,6 +52,7 @@ class StoreDeleted(object):
     def __reduce__(self):
         return "deleted"
 
+
 deleted = StoreDeleted()
 
 
@@ -83,6 +83,12 @@ def get_store_module(name):
     return sys.modules[name]
 
 
+from renpy.pydict import DictItems, find_changes
+
+EMPTY_DICT = { }
+EMPTY_SET = set()
+
+
 class StoreDict(dict):
     """
     This class represents the dictionary of a store module. It logs
@@ -96,7 +102,7 @@ class StoreDict(dict):
 
         # The value of this dictionary at the start of the current
         # rollback period (when begin() was last called).
-        self.old = { }
+        self.old = DictItems(self)
 
         # The set of variables in this StoreDict that changed since the
         # end of the init phase.
@@ -107,43 +113,59 @@ class StoreDict(dict):
         Called to reset this to its initial conditions.
         """
 
-        self.old = { }
         self.ever_been_changed = set()
         self.clear()
+        self.old = DictItems(self)
 
     def begin(self):
         """
         Called to mark the start of a rollback period.
         """
 
-        self.old = dict(self)
+        self.old = DictItems(self)
 
-    def get_changes(self):
+    def get_changes(self, cycle):
         """
         For every key that has changed since begin() was called, returns a
         dictionary mapping the key to its value when begin was called, or
         deleted if it did not exist when begin was called.
 
-        As a side-effect, updates self.ever_been_changed.
+        As a side-effect, updates self.ever_been_changed, and returns the
+        changes to ever_been_changed as well.
+
+        `cycle`
+            If true, this cycles the old changes to the new changes. If
+            False, does not.
         """
 
-        rv = { }
+        new = DictItems(self)
+        rv = find_changes(self.old, new, deleted)
 
-        for k in self:
-            if k not in self.old:
-                rv[k] = deleted
+        if cycle:
+            self.old = new
 
-        for k, v in self.old.iteritems():
+        if rv is None:
+            return EMPTY_DICT, EMPTY_SET
 
-            new_v = self.get(k, deleted)
+        delta_ebc = set()
 
-            if new_v is not v:
-                rv[k] = v
+        if cycle:
 
-        for k in rv:
-            self.ever_been_changed.add(k)
+            for k in rv:
+                if k not in self.ever_been_changed:
+                    self.ever_been_changed.add(k)
+                    delta_ebc.add(k)
 
-        return rv
+        return rv, delta_ebc
+
+
+def begin_stores():
+    """
+    Calls .begin on every store dict.
+    """
+
+    for sd in store_dicts.itervalues():
+        sd.begin()
 
 
 # A map from the name of a store dict to the corresponding StoreDict object.
@@ -217,10 +239,16 @@ class StoreBackup():
         # The contents of ever_been_changed for each store.
         self.ever_been_changed = { }
 
-        for k, v in store_dicts.iteritems():
-            self.store[k] = dict(v)
-            self.old[k] = dict(v.old)
-            self.ever_been_changed[k] = set(v.ever_been_changed)
+        for k in store_dicts:
+            self.backup_one(k)
+
+    def backup_one(self, name):
+
+        d = store_dicts[name]
+
+        self.store[name] = dict(d)
+        self.old[name] = d.old.as_dict()
+        self.ever_been_changed[name] = set(d.ever_been_changed)
 
     def restore_one(self, name):
         sd = store_dicts[name]
@@ -228,8 +256,7 @@ class StoreBackup():
         sd.clear()
         sd.update(self.store[name])
 
-        sd.old.clear()
-        sd.old.update(self.old[name])
+        sd.old = DictItems(self.old[name])
 
         sd.ever_been_changed.clear()
         sd.ever_been_changed.update(self.ever_been_changed[name])
@@ -252,8 +279,8 @@ def make_clean_stores():
 
     for _k, v in store_dicts.iteritems():
 
-        v.old.clear()
         v.ever_been_changed.clear()
+        v.begin()
 
     clean_store_backup = StoreBackup()
 
@@ -322,7 +349,7 @@ def reached(obj, reachable, wait):
     if idobj in reachable:
         return
 
-    if isinstance(obj, (NoRollback, file)):
+    if isinstance(obj, (NoRollback, real_file)):  # @UndefinedVariable
         reachable[idobj] = 0
         return
 
@@ -404,6 +431,18 @@ class WrapNode(ast.NodeTransformer):
             starargs=None,
             kwargs=None)
 
+    def visit_Set(self, n):
+
+        return ast.Call(
+            func=ast.Name(
+                id="__renpy__set__",
+                ctx=ast.Load()
+                ),
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
+
     def visit_ListComp(self, n):
         return ast.Call(
             func=ast.Name(
@@ -452,21 +491,27 @@ class WrapNode(ast.NodeTransformer):
             starargs=None,
             kwargs=None)
 
+
 wrap_node = WrapNode()
 
 
-def set_filename(filename, offset, tree):
-    """Set the filename attribute to filename on every node in tree"""
-    worklist = [tree]
-    while worklist:
-        node = worklist.pop(0)
-        node.filename = filename
+def wrap_hide(tree):
+    """
+    Wraps code inside a python hide or python early hide block inside a
+    function, so it gets its own scope that works the way Python expects
+    it to.
+    """
 
-        lineno = getattr(node, 'lineno', None)
-        if lineno is not None:
-            node.lineno = lineno + offset
+    hide = ast.parse("""\
+def _execute_python_hide(): pass;
+_execute_python_hide()
+""")
 
-        worklist.extend(node.getChildNodes())
+    for i in ast.walk(hide):
+        ast.copy_location(i, hide.body[0])
+
+    hide.body[0].body = tree.body
+    tree.body = hide.body
 
 
 unicode_re = re.compile(ur'[\u0080-\uffff]')
@@ -517,7 +562,14 @@ new_compile_flags = (  old_compile_flags
                        )
 
 
-def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
+# A cache for the results of py_compile.
+py_compile_cache = { }
+
+# An old version of the same, that's preserved across reloads.
+old_py_compile_cache = { }
+
+
+def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=True):
     """
     Compiles the given source code using the supplied codegenerator.
     Lists, List Comprehensions, and Dictionaries are wrapped when
@@ -543,6 +595,9 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         that would be used.
     """
 
+    if ast_node:
+        cache = False
+
     if isinstance(source, ast.Module):
         return compile(source, filename, mode)
 
@@ -550,6 +605,26 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         filename = source.filename
         lineno = source.linenumber
 
+    if cache:
+        key = (lineno, filename, unicode(source), mode, renpy.script.MAGIC)
+
+        rv = py_compile_cache.get(key, None)
+        if rv is not None:
+            return rv
+
+        rv = old_py_compile_cache.get(key, None)
+        if rv is not None:
+            old_py_compile_cache[key] = rv
+            return rv
+
+        bytecode = renpy.game.script.bytecode_oldcache.get(key, None)
+        if bytecode is not None:
+
+            renpy.game.script.bytecode_newcache[key] = bytecode
+            rv = marshal.loads(bytecode)
+            py_compile_cache[key] = rv
+            return rv
+
     source = unicode(source)
     source = source.replace("\r", "")
     source = escape_unicode(source)
@@ -557,15 +632,23 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
     try:
         line_offset = lineno - 1
 
+        if mode == "hide":
+            py_mode = "exec"
+        else:
+            py_mode = mode
+
         try:
             flags = new_compile_flags
-            tree = compile(source, filename, mode, ast.PyCF_ONLY_AST | flags, 1)
+            tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
         except:
             flags = old_compile_flags
-            tree = compile(source, filename, mode, ast.PyCF_ONLY_AST | flags, 1)
+            tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
 
         tree = wrap_node.visit(tree)
 
+        if mode == "hide":
+            wrap_hide(tree)
+
         ast.fix_missing_locations(tree)
         ast.increment_lineno(tree, lineno - 1)
 
@@ -574,7 +657,14 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         if ast_node:
             return tree.body
 
-        return compile(tree, filename, mode, flags, 1)
+        rv = compile(tree, filename, py_mode, flags, 1)
+
+        if cache:
+            py_compile_cache[key] = rv
+            renpy.game.script.bytecode_newcache[key] = marshal.dumps(rv)
+            renpy.game.script.bytecode_dirty = True
+
+        return rv
 
     except SyntaxError, e:
 
@@ -585,13 +675,18 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
 
 
 def py_compile_exec_bytecode(source, **kwargs):
-    code = py_compile(source, 'exec', **kwargs)
+    code = py_compile(source, 'exec', cache=False, **kwargs)
+    return marshal.dumps(code)
+
+
+def py_compile_hide_bytecode(source, **kwargs):
+    code = py_compile(source, 'hide', cache=False, **kwargs)
     return marshal.dumps(code)
 
 
 def py_compile_eval_bytecode(source, **kwargs):
     source = source.strip()
-    code = py_compile(source, 'eval', **kwargs)
+    code = py_compile(source, 'eval', cache=False, **kwargs)
     return marshal.dumps(code)
 
 
@@ -677,10 +772,10 @@ class CompressedList(object):
             old_end += 1
 
         # Now that we have this, we can put together the object.
-        self.pre = old[0:old_start]
+        self.pre = list.__getslice__(old, 0, old_start)
         self.start = new_start
         self.end = new_end
-        self.post = old[old_end:]
+        self.post = list.__getslice__(old, old_end, len_old)
 
     def decompress(self, new):
         return self.pre + new[self.start:self.end] + self.post
@@ -718,16 +813,21 @@ class RevertableList(list):
 
     def wrapper(method):  # E0213 @NoSelf
         def newmethod(*args, **kwargs):
-            return RevertableList(method(*args, **kwargs))  # E1102
+            l = method(*args, **kwargs)
+            return RevertableList(l)
 
         return newmethod
 
     __add__ = wrapper(list.__add__)
     __getslice__ = wrapper(list.__getslice__)
-    __mul__ = wrapper(list.__mul__)
-    __rmul__ = wrapper(list.__rmul__)
 
-    del wrapper
+    def __mul__(self, other):
+        if not isinstance(other, int):
+            raise TypeError("can't multiply sequence by non-int of type '{}'.".format(type(other).__name__))
+
+        return RevertableList(list.__mul__(self, other))
+
+    __rmul__ = __mul__
 
     def _clean(self):
         """
@@ -791,6 +891,7 @@ class RevertableDict(dict):
     pop = mutator(dict.pop)
     popitem = mutator(dict.popitem)
     setdefault = mutator(dict.setdefault)
+    update = mutator(dict.update)
 
     def list_wrapper(method):  # E0213 @NoSelf
         def newmethod(*args, **kwargs):
@@ -822,7 +923,21 @@ class RevertableDict(dict):
             self[k] = v
 
 
-class RevertableSet(sets.Set):
+class RevertableSet(set):
+
+    def __setstate__(self, state):
+        if isinstance(state, tuple):
+            self.update(state[0].keys())
+        else:
+            self.update(state)
+
+    def __getstate__(self):
+        rv = ({ i : True for i in self}, )
+        return rv
+
+    # Required to ensure that getstate and setstate are called.
+    __reduce__ = object.__reduce__
+    __reduce_ex__ = object.__reduce_ex__
 
     def __init__(self, *args):
         log = renpy.game.log
@@ -830,44 +945,42 @@ class RevertableSet(sets.Set):
         if log is not None:
             log.mutated[id(self)] = None
 
-        sets.Set.__init__(self, *args)
+        set.__init__(self, *args)
 
-    __iand__ = mutator(sets.Set.__iand__)
-    __ior__ = mutator(sets.Set.__ior__)
-    __isub__ = mutator(sets.Set.__isub__)
-    __ixor__ = mutator(sets.Set.__ixor__)
-    add = mutator(sets.Set.add)
-    clear = mutator(sets.Set.clear)
-    difference_update = mutator(sets.Set.difference_update)
-    discard = mutator(sets.Set.discard)
-    intersection_update = mutator(sets.Set.intersection_update)
-    pop = mutator(sets.Set.pop)
-    remove = mutator(sets.Set.remove)
-    symmetric_difference_update = mutator(sets.Set.symmetric_difference_update)
-    union_update = mutator(sets.Set.union_update)
-    update = mutator(sets.Set.update)
+    __iand__ = mutator(set.__iand__)
+    __ior__ = mutator(set.__ior__)
+    __isub__ = mutator(set.__isub__)
+    __ixor__ = mutator(set.__ixor__)
+    add = mutator(set.add)
+    clear = mutator(set.clear)
+    difference_update = mutator(set.difference_update)
+    discard = mutator(set.discard)
+    intersection_update = mutator(set.intersection_update)
+    pop = mutator(set.pop)
+    remove = mutator(set.remove)
+    symmetric_difference_update = mutator(set.symmetric_difference_update)
+    union_update = mutator(set.update)
+    update = mutator(set.update)
 
-    def wrapper(method):  # E0213 @NoSelf
+    def wrapper(method):  # @NoSelf
         def newmethod(*args, **kwargs):
-            rv = method(*args, **kwargs)  # E1102
-            if isinstance(rv, sets.Set):
+            rv = method(*args, **kwargs)
+            if isinstance(rv, (set, frozenset)):
                 return RevertableSet(rv)
             else:
                 return rv
 
         return newmethod
 
-    __and__ = wrapper(sets.Set.__and__)
-    __copy__ = wrapper(sets.Set.__copy__)
-    __deepcopy__ = wrapper(sets.Set.__deepcopy__)
-    __sub__ = wrapper(sets.Set.__sub__)
-    __xor__ = wrapper(sets.Set.__xor__)
-    __or__ = wrapper(sets.Set.__or__)
-    copy = wrapper(sets.Set.copy)
-    difference = wrapper(sets.Set.difference)
-    intersection = wrapper(sets.Set.intersection)
-    symmetric_difference = wrapper(sets.Set.symmetric_difference)
-    union = wrapper(sets.Set.union)
+    __and__ = wrapper(set.__and__)
+    __sub__ = wrapper(set.__sub__)
+    __xor__ = wrapper(set.__xor__)
+    __or__ = wrapper(set.__or__)
+    copy = wrapper(set.copy)
+    difference = wrapper(set.difference)
+    intersection = wrapper(set.intersection)
+    symmetric_difference = wrapper(set.symmetric_difference)
+    union = wrapper(set.union)
 
     del wrapper
 
@@ -878,8 +991,8 @@ class RevertableSet(sets.Set):
         return clean
 
     def _rollback(self, compressed):
-        sets.Set.clear(self)
-        sets.Set.update(self, compressed)
+        set.clear(self)
+        set.update(self, compressed)
 
 
 class RevertableObject(object):
@@ -893,6 +1006,10 @@ class RevertableObject(object):
 
         return self
 
+    def __init__(self, *args, **kwargs):
+        if (args or kwargs) and renpy.config.developer:
+            raise TypeError("object() takes no parameters.")
+
     def __setattr__(self, attr, value):
         object.__setattr__(self, attr, value)
 
@@ -962,6 +1079,7 @@ class DetRandom(random.Random):
         new.seed(seed)
         return new
 
+
 rng = DetRandom()
 
 # This is the code that actually handles the logging and managing
@@ -1000,7 +1118,7 @@ class Rollback(renpy.object.Object):
     execution of this element.
     """
 
-    __version__ = 4
+    __version__ = 5
 
     identifier = None
 
@@ -1009,6 +1127,7 @@ class Rollback(renpy.object.Object):
         super(Rollback, self).__init__()
 
         self.context = renpy.game.context().rollback_copy()
+
         self.objects = [ ]
         self.purged = False
         self.random = [ ]
@@ -1017,6 +1136,10 @@ class Rollback(renpy.object.Object):
         # A map of maps name -> (variable -> value)
         self.stores = { }
 
+        # A map from store name to the changes to ever_been_changed that
+        # need to be reverted.
+        self.delta_ebc = { }
+
         # If true, we retain the data in this rollback when a load occurs.
         self.retain_after_load = False
 
@@ -1052,6 +1175,9 @@ class Rollback(renpy.object.Object):
         if version < 4:
             self.hard_checkpoint = self.checkpoint
 
+        if version < 5:
+            self.delta_ebc = { }
+
     def purge_unreachable(self, reachable, wait):
         """
         Adds objects that are reachable from the store of this
@@ -1092,8 +1218,7 @@ class Rollback(renpy.object.Object):
                 reached(rb, reachable, wait)
             else:
                 if renpy.config.debug:
-                    print(("Removing unreachable:", o))
-
+                    print("Removing unreachable:", o, file=renpy.log.real_stdout)
                     pass
 
         self.objects = new_objects
@@ -1113,7 +1238,7 @@ class Rollback(renpy.object.Object):
         for name, changes in self.stores.iteritems():
             store = store_dicts.get(name, None)
             if store is None:
-                return
+                continue
 
             for name, value in changes.iteritems():
                 if value is deleted:
@@ -1122,6 +1247,14 @@ class Rollback(renpy.object.Object):
                 else:
                     store[name] = value
 
+        for name, changes in self.delta_ebc.iteritems():
+
+            store = store_dicts.get(name, None)
+            if store is None:
+                continue
+
+            store.ever_been_changed -= changes
+
         rng.pushback(self.random)
 
         renpy.game.contexts.pop()
@@ -1184,6 +1317,10 @@ class RollbackLog(renpy.object.Object):
         # on load.
         self.retain_after_load_flag = False
 
+        # Has there been an interaction since the last time this log was
+        # reset?
+        self.did_interaction = True
+
     def after_setstate(self):
         self.mutated = { }
         self.rolled_forward = False
@@ -1210,7 +1347,7 @@ class RollbackLog(renpy.object.Object):
 
                 self.rollback_limit = nrbl
 
-    def begin(self):
+    def begin(self, force=False):
         """
         Called before a node begins executing, to indicate that the
         state needs to be saved for rollbacking.
@@ -1223,12 +1360,30 @@ class RollbackLog(renpy.object.Object):
         if not context.rollback:
             return
 
-        # If the transient scene list is not empty, then we do
-        # not begin a new rollback, as the TSL will be purged
-        # after a rollback is complete.
-        if not context.scene_lists.transient_is_empty():
+        # We only begin a checkpoint if the previous statement reached a checkpoint,
+        # or an interaction took place. (Or we're forced.)
+        ignore = True
+
+        if force:
+            ignore = False
+        elif self.did_interaction:
+            ignore = False
+        elif self.current is not None:
+            if self.current.checkpoint:
+                ignore = False
+            elif self.current.retain_after_load:
+                ignore = False
+
+        if ignore:
             return
 
+        self.did_interaction = False
+
+        if self.current is not None:
+            self.complete(True)
+        else:
+            begin_stores()
+
         # If the log is too long, prune it.
         if len(self.log) > renpy.config.rollback_length:
             self.log = self.log[-renpy.config.rollback_length:]
@@ -1258,23 +1413,22 @@ class RollbackLog(renpy.object.Object):
 
         self.rolled_forward = False
 
-        # Reset the point that changes are relative to.
-        for sd in store_dicts.itervalues():
-            sd.begin()
-
-    def complete(self):
+    def complete(self, begin=False):
         """
         Called after a node is finished executing, before a save
         begins, or right before a rollback is attempted. This may be
         called more than once between calls to begin, and should always
         be called after an update to the store but before a rollback
         occurs.
+
+        `begin`
+            Should be true if called from begin().
         """
 
         # Update self.current.stores with the changes from each store.
         # Also updates .ever_been_changed.
         for name, sd in store_dicts.iteritems():
-            self.current.stores[name] = sd.get_changes()
+            self.current.stores[name], self.current.delta_ebc[name] = sd.get_changes(begin)
 
         # Update the list of mutated objects and what we need to do to
         # restore them.
@@ -1322,6 +1476,9 @@ class RollbackLog(renpy.object.Object):
                 else:
                     rv[store_name + "." + name] = deleted
 
+        for i in reversed(renpy.game.contexts[1:]):
+            i.pop_dynamic_roots(rv)
+
         return rv
 
     def purge_unreachable(self, roots, wait=None):
@@ -1409,8 +1566,8 @@ class RollbackLog(renpy.object.Object):
                 fwd_name, fwd_data = self.forward[0]
 
                 if (self.current.context.current == fwd_name
-                            and data == fwd_data
-                            and (keep_rollback or self.rolled_forward)
+                        and data == fwd_data
+                        and (keep_rollback or self.rolled_forward)
                         ):
                     self.forward.pop(0)
                 else:
@@ -1441,8 +1598,12 @@ class RollbackLog(renpy.object.Object):
         when the game is loaded.
         """
 
+        if renpy.display.predict.predicting:
+            return
+
         self.retain_after_load_flag = True
         self.current.retain_after_load = True
+        renpy.game.context().force_checkpoint = True
 
     def fix_rollback(self):
         if not self.rollback_is_fixed and len(self.log) > 1:
@@ -1595,13 +1756,21 @@ class RollbackLog(renpy.object.Object):
 
         renpy.game.contexts.extend(other_contexts)
 
-        if force_checkpoint:
-            renpy.game.context().force_checkpoint = True
+        begin_stores()
 
         # Restart the context or the top context.
         if replace_context:
+
+            if force_checkpoint:
+                renpy.game.contexts[0].force_checkpoint = True
+
             raise renpy.game.RestartTopContext(label)
+
         else:
+
+            if force_checkpoint:
+                renpy.game.context().force_checkpoint = True
+
             raise renpy.game.RestartContext(label)
 
     def freeze(self, wait=None):
@@ -1613,7 +1782,7 @@ class RollbackLog(renpy.object.Object):
         """
 
         # Purge unreachable objects, so we don't save them.
-        self.complete()
+        self.complete(False)
         roots = self.get_roots()
         self.purge_unreachable(roots, wait=wait)
 
@@ -1746,6 +1915,7 @@ def py_eval_bytecode(bytecode, globals=None, locals=None):  # @ReservedAssignmen
 def py_eval(code, globals=None, locals=None):  # @ReservedAssignment
     if isinstance(code, basestring):
         code = py_compile(code, 'eval')
+
     return py_eval_bytecode(code, globals, locals)
 
 
@@ -1796,6 +1966,22 @@ def method_pickle(method):
 def method_unpickle(obj, name):
     return getattr(obj, name)
 
+# Code for pickling modules.
+
+
+def module_pickle(module):
+    if renpy.config.developer:
+        raise Exception("Could not pickle {!r}.".format(module))
+
+    return module_unpickle, (module.__name__,)
+
+
+def module_unpickle(name):
+    return __import__(name)
+
+
 import copy_reg
 import types
-copy_reg.pickle(types.MethodType, method_pickle, method_unpickle)
+
+copy_reg.pickle(types.MethodType, method_pickle)
+copy_reg.pickle(types.ModuleType, module_pickle)
